import PmedConnect.ParseFields as pf
import PmedConnect.config as config
import calendar

class Parser(object):
  available_parse_paths = config.PUBMED_PATHS

  # This is for translating abbreviated month names to numbers.
  months_rdict = {v: str(k) for k, v in enumerate(calendar.month_abbr)}    

  def __init__(self, fields, extra_parse_functions = None):
    self.build_parse_functions()
    # self.add_extra_parse_functions(extra_parse_functions)

    field_validator = pf.ParseFields(self.parse_functions)
    
    self.fields = field_validator.validate(fields)

  def extract_from_docs(self, docs):
    parsed_docs = []

    for doc in docs:
      parsed_docs.append(self.extract_fields(doc))

    return parsed_docs

  def extract_fields(self, doc):
    """Extracts the required fields from the documents,
    where each document is an Entrez dictionary"""    
    parsed_doc = {}

    parsed_doc['doc_type'] = self.determine_doc_type(doc)

    for field in self.fields:
      parse_func = self.parse_functions.get(field)
      
      parsed_doc[field] = parse_func(doc)

    return parsed_doc

  def determine_doc_type(self, doc):
    for path in self.available_parse_paths['path_determine_type']['paths']:
      try:
        data = doc

        for node in path:
          data = data[node]

        if node == 'PublicationTypeList':
          return str(data[0])
        
        return node
      except KeyError:
        continue

    return 'Other'


  def parse_abstract(self, abs_str):
    """Abstracts may contain multiple elements, convert to string and return joined string"""
    return ' '.join(abs_str).encode('utf-8')

  def format_ddate(self, ddate):
    """Turn a date dictionary into an ISO-type string (YYYY-MM-DD)."""
    year = ddate['Year']
    month = ddate['Month']
    day = ddate['Day']
    
    if not month.isdigit():
      month = self.months_rdict.get(month, None)
      
      if not month:
        return None

    return "%s-%s-%s" % (year, month.zfill(2), day.zfill(2))

  def extract_id_factory(self, idtype):
    """Extract an ID from Entrez XML output."""
    def extract_id(doc_data, key = 'PubmedData'):
      found = None

      try:
        for articleid in doc_data[key]['ArticleIdList']:
          if articleid.attributes['IdType'].lower() == idtype:
            found = str(articleid)
      except (KeyError):
        if key is 'PubmedData':
          return extract_id(doc_data, 'PubmedBookData')

      if found is None and 'path_' + idtype in self.available_parse_paths:
        found = self.extract_path_factory(idtype)(doc_data)

        if found is not None:
          found = found[0]

      return found
    
    return extract_id

  def extract_date_factory(self, datetype):
    def extract_date(doc_data, key = 'PubmedData'):
      try:
        for date in doc_data[key]['History']:
          if date.attributes['PubStatus'].lower() == datetype:
            return self.format_ddate(date)
      except (KeyError):
        if key is 'PubmedData':
          return extract_date(doc_data, 'PubmedBookData')
    
    return extract_date

  def extract_path_factory(self, field, rdict = None, fmt = None):
    def extract_path(doc_data):
      for path in self.available_parse_paths['path_' + field]['paths']:
        data = doc_data

        try:
          for node in path:
            data = data[node]

          if len(data) < 1:
            continue

          if rdict != None:
            data = rdict[data]

          if fmt != None:
            data = fmt(data)

          return data
        except (KeyError, TypeError):
          continue
          
      return None

    return extract_path

  def extract_keywords_factory(self):
    def extract_keywords(doc_data):
      for path in self.available_parse_paths['path_keywords']['paths']:
        try:
          data = doc_data

          for node in path:
            data = data[node]

          keywords = []
          
          try:
            for keyword in data[0]:
              keywords.append(str(keyword))
          except IndexError:
            return keywords

          return keywords
        except (KeyError, TypeError):
          continue

      return None

    return extract_keywords

  def add_extra_parse_functions(self, parse_functions):
    raise NotImplementedError
    # if parse_functions is None:
    #   return

    # for name, parse_function in parse_functions.items():
    #   self.parse_functions[name] = parse_function

  def build_parse_functions(self):
    self.parse_functions = {
      'pmid': self.extract_id_factory('pubmed'),
      'doi': self.extract_id_factory('doi'),

      'date_pubmed_published': self.extract_date_factory('pubmed'),
      'date_medline_published': self.extract_date_factory('medline'),

      'title': self.extract_path_factory('title'),
      'abstract': self.extract_path_factory('abstract', fmt = self.parse_abstract),
      'pub_year': self.extract_path_factory('pub_year'),
      'pub_month': self.extract_path_factory('pub_month', rdict = self.months_rdict),
      'pub_day': self.extract_path_factory('pub_day'),
      'vol': self.extract_path_factory('vol'),
      'pages': self.extract_path_factory('pages'),
      'date_pubmed_created': self.extract_path_factory('date_pubmed_created', fmt = self.format_ddate),
      'date_pubmed_updated': self.extract_path_factory('date_pubmed_updated', fmt = self.format_ddate),

      'journal_title': self.extract_path_factory('journal_title'),
      'journal_iso': self.extract_path_factory('journal_iso'),
      'journal_issn': self.extract_path_factory('journal_issn', fmt = str),

      'keywords': self.extract_keywords_factory()
    }


import PmedConnect.config as config, PmedConnect.ParseFunctions as xpf, PmedConnect.ReadPMC as xr
import math, progressbar, time

from datetime import date
from Bio import Entrez

class PubmedAPI(object):
  """Searches and fetches data from Pubmed using the Entrez module from Biopython."""

  xml_reader = None
  search_progressbar_obj = None

  mindate = None
  maxdate = None
  
  retall = True
  retnum = config.RESULTS_PER_SEARCH_REQUEST
  retmax = config.RESULTS_PER_SEARCH_REQUEST
  
  # Summary variables, used to print at the end of a search
  total_results = 0
  found_results = 0
  query = ''
  interpreted_query = ''
  rounds_made = 0

  # Constants
  converted_ids = False
  db = 'pubmed'
  retmode = 'xml'
  datetype = 'pdat'
  fetch_block = config.RESULTS_PER_FETCH_REQUEST

  def __init__(self, email, fields = None):
    # Entrez requires an email address.
    Entrez.email = email

    self.parser = xpf.Parser(fields)

    # Set default search database
    self.db = 'pubmed'

    self.pmid_extractor = self.parser.extract_id_factory('pubmed')

  def set_search_database(self, db):
    available_dbs = ['pubmed', 'pmc']

    if db not in available_dbs:
      raise ValueError('This database is not supported')

    self.db = db

    if self.db is 'pmc':
      self.xml_reader = xr.ReadPMC()

  def set_search_date(self, mindate, maxdate = None):
    """Restricts searches to the specified date range (format as YYYY/MM/DD, YYYY/MM, or YYYY)"""

    if maxdate == None:
      today = date.fromtimestamp(time.time())

      maxdate = today.strftime('%Y/%m/%d')

    self.mindate = mindate
    self.maxdate = maxdate

  def reset_search_data(self):
    self.mindate = None
    self.maxdate = None

  def set_retrieve_params(self, num_retrieve, num_retrieve_per_round):
    """Set the number of items to retrieve from the Entrez API each search call.
    If the number of retrieved items in a search does not satisfy the limit
    AND there are more items to retrieve from the API, the search will
    iterate using retstart."""
    self.retall = True
    self.retmax = config.RESULTS_PER_SEARCH_REQUEST

    # Check whether num_retrieve_per_round is provided
    if num_retrieve_per_round is not None:
      # Check against the Entrez API limits
      if num_retrieve_per_round > config.ENTREZ_MAX_RETRIEVE:
        raise ValueError('Number of items to retrieve per round lies outside the maximum of the Entrez API')

      self.retmax = num_retrieve_per_round

    # Check whether num_retrieve is provided
    if num_retrieve is not None:
      self.retall = False
      self.retnum = num_retrieve

      # Adjust the maximum number of items to retrieve per search.
      # If the number to retrieve is lower than the current maximum
      # number of results per request, set the retmax to num_retrieve.
      if num_retrieve < self.retmax:
        self.retmax = num_retrieve

  def check_search_error(self, search_results):
    """Checks whether the search results contain an error, in which case an exception is raised"""
    if 'WarningList' in search_results:
      warnings = '; '.join(search_results['WarningList']['OutputMessage'])

      # Regular error when no results are returned for a search
      if 'No items found' in warnings and config.IGNORE_NO_ITEMS:
        return

      raise Exception('Entrez error: %s' % (warnings))

  def update_search_summary(self, search_results, num_ids):
    self.total_results = int(search_results['Count'])
    self.found_results = num_ids
    self.interpreted_query = search_results['QueryTranslation']
    self.rounds_made = self.rounds_made + 1

  def attach_search_summary(self, pmids):
    return {
      'pmids': pmids,
      'summary': {
        'translated_query': self.interpreted_query,
        'total_found': self.total_results,
        'retrieved': self.found_results,
        'rounds_made': self.rounds_made,
      }
    }

  def update_search_progressbar(self, initialise, max_results, current_results):
    """Initialises and updates the progressbar during searches"""
    if config.SILENT is True:
      return

    if initialise == 0:
      bar_max = self.retnum

      if self.retnum == 0 or self.retnum > max_results:
        bar_max = max_results

      
      self.search_progressbar_obj = progressbar.ProgressBar(max_value = bar_max)
      
    self.search_progressbar_obj.update(current_results)

  def get_search_params(self, query, retstart):
    """Creates a dictionary of parameters to pass into the Entrez.esearch function"""
    func_params = dict(db = self.db, term = query, retmode = self.retmode, retstart = retstart, retmax = self.retmax)

    if self.mindate is not None:
      func_params['datetype'] = self.datetype
      func_params['mindate'] = self.mindate
      func_params['maxdate'] = self.maxdate

    return func_params

  def get_search_total_count(self, results):
    """Returns the total number of results for the search"""
    self.check_search_error(results)

    return int(results['Count'])

  def get_search_result_ids(self, existing_ids, results):
    """Extracts the PMID list from the search results, cuts the list
    of IDs if it is bigger than the maximum number of results to
    return (i.e. self.retnum)"""
    self.check_search_error(results)
    
    ids = existing_ids + results['IdList']

    # Cut IDs to size of maximum to retrieve
    if len(ids) > self.retnum:
      ids = ids[0:self.retnum]

    return ids

  def search_round(self, query, retstart = 0, found_ids = []):
    """Searches for the provided parameters, iteratively calls itself
    if there are more results to retrieve."""
    search_params = self.get_search_params(query, retstart)
    search_results = Entrez.read(Entrez.esearch(**search_params))

    total_count = self.get_search_total_count(search_results)
    found_ids = self.get_search_result_ids(found_ids, search_results)

    self.update_search_progressbar(retstart, total_count, len(found_ids))

    # Check for empty searches
    if len(found_ids) == 0 or total_count == 0:
      return found_ids
    
    # Check if another search is possible and/or necessary
    if (self.retall == True or len(found_ids) < self.retnum) and total_count > retstart:
      retstart = retstart + self.retmax

      found_ids = self.search_round(query, retstart, found_ids)

    self.update_search_summary(search_results, len(found_ids))

    return found_ids

  def search(self, query, num_retrieve = None, num_retrieve_per_round = None):
    """Uses the esearch function to search in the Entrez API"""
    self.set_retrieve_params(num_retrieve, num_retrieve_per_round)

    pmids = self.search_round(query)
    
    return self.attach_search_summary(pmids)

  def read_results(self, fetch_results, pmids):
    # For the reasoning behind this function read the comment
    # below in the function parse_results.

    if self.db is 'pubmed' or self.converted_ids is True:
      read_results = Entrez.read(fetch_results)

      articles = read_results['PubmedArticle']
      books = read_results['PubmedBookArticle']

      # Reorder the documents by the original pmids list.
      documents = articles + books

      fetched_pmid_ids = []

      for document in documents:
        fetched_pmid_ids.append(self.pmid_extractor(document))

      pmid_copy = pmids
      new_keys = []

      for item in fetched_pmid_ids:
        new_key = pmid_copy.index(item)
        
        pmid_copy[new_key] = ''

        new_keys.append(new_key)

      documents[:] = [a for b,a in sorted(zip(new_keys, documents))]

      return documents
    
    if self.db is 'pmc':
      return fetch_results

  def fetch_round(self, pmids):
    """Fetch and parse the XML for a list of PMIDs."""
    fetch_db = self.db

    if self.converted_ids is True:
      fetch_db = 'pubmed'

    func_params = dict(db = fetch_db, id = pmids, retmode = self.retmode)

    fetch_results = Entrez.efetch(**func_params)
    
    return self.read_results(fetch_results, pmids)

  def parse_results(self, doc_list):
    if self.db is 'pubmed' or self.converted_ids is True:
      self.converted_ids = False

      return self.parser.extract_from_docs(doc_list)

    if self.db is 'pmc':
      # PMC searches yield PMC IDs. The Bio.Entrez does
      # not handle PMC XML. Therefore, the first round of 
      # fetching is used to get the pmids. Run the pmids
      # through another round of fetching to build the
      # Bio.Entrez data format.
      # This is not efficient but much easier than building
      # the data format from the PMC XML.
      pmids = self.xml_reader.get_pmids(doc_list)

      self.converted_ids = True

      return self.fetch(pmids)

  def fetch(self, pmid):
    """Breaks Entrez fetching process into blocks,
    fetches documents in chunks of fetch_block."""
    number_of_rounds = int(math.floor(len(pmid) / self.fetch_block + (len(pmid) % self.fetch_block > 0)))

    self.set_retrieve_params(len(pmid), self.fetch_block)
    
    doc_list = []

    if number_of_rounds > 1:
      print('Will fetch in %i rounds with %i articles per round.' % (number_of_rounds, self.fetch_block))

      for i in range(number_of_rounds):
        start = i * self.fetch_block
        finish = (i + 1) * self.fetch_block

        self.update_search_progressbar(i, len(pmid), start)
        
        doc_list = doc_list + self.fetch_round(pmid[start:finish])

      print('\n')
    else:
      print('Will fetch %i articles.' % len(pmid))

      doc_list = self.fetch_round(pmid)
    
    parsed_results = self.parse_results(doc_list)

    return parsed_results


import PmedConnect.config as config

class ParseFields(object):
  def __init__(self, default_parse_functions):
    self.available_fields = []
    self.add_functions(default_parse_functions)

  def add_functions(self, parse_functions):
    for name, parse_function in parse_functions.items():
      self.available_fields.append(name)

  def validate(self, fields = None):
    # If single field is passed, convert to list
    if type(fields) is not list and type(fields) is str:
      fields = [fields]

    # If nothing is passed return the default fields
    if fields is None:
      return self.available_fields
    
    # Get the differences between the fields passed into the function
    # and the list of available fields from the config
    unavailable_fields = set(fields) - set(self.available_fields)

    # Raise error when a field is not available
    if len(unavailable_fields) > 0:
      raise ValueError('Following fields do not have a parsing function: ' + ', '.join(map(str, unavailable_fields)))

    return fields


ENTREZ_MAX_RETRIEVE = 100000

RESULTS_PER_SEARCH_REQUEST = 10000
RESULTS_PER_FETCH_REQUEST = 10000

AVAILABLE_FIELDS = [
  'pmid', 'doi', 'title', 'abstract',
  'pub_year', 'pub_month', 'pub_day',
  'journal_title', 'journal_iso', 'journal_issn',
  'keywords',
  'date_pubmed_created', 'date_pubmed_updated',
  'date_pubmed_published', 'date_medline_published'
]

RETRIEVE_FIELDS = AVAILABLE_FIELDS

PUBMED_PATHS = {
  'path_determine_type': {'paths': [['MedlineCitation', 'Article', 'PublicationTypeList'], ['MedlineCitation', 'Article'], ['BookDocument', 'Book']]},

  'path_doi': {'paths': [['MedlineCitation', 'Article', 'ELocationID']]},
  'path_title': {'paths': [['MedlineCitation', 'Article', 'ArticleTitle'], ['BookDocument', 'ArticleTitle'], ['BookDocument', 'Book', 'BookTitle']]},
  'path_abstract': {'paths': [['MedlineCitation', 'Article', 'Abstract', 'AbstractText'], ['BookDocument', 'Abstract', 'AbstractText']]},
  'path_vol': {'paths': [['MedlineCitation', 'Article', 'Journal', 'JournalIssue', 'Volume']]},
  'path_pages': {'paths': [['MedlineCitation', 'Article', 'Pagination', 'MedlinePgn']]},
  'path_pub_year': {'paths': [['MedlineCitation', 'Article', 'Journal', 'JournalIssue', 'PubDate', 'Year'], ['PubmedBookData', 'History', 0, 'Year']]},
  'path_pub_month': {'paths': [['MedlineCitation', 'Article', 'Journal', 'JournalIssue', 'PubDate', 'Month'], ['PubmedBookData', 'History', 0, 'Month']]},
  'path_pub_day': {'paths': [['MedlineCitation', 'Article', 'Journal', 'JournalIssue', 'PubDate', 'Day'], ['PubmedBookData', 'History', 0, 'Day']]},
  'path_date_pubmed_created': {'paths': [['MedlineCitation', 'DateCreated'], ['BookDocument', 'ContributionDate']]},
  'path_date_pubmed_updated': {'paths': [['MedlineCitation', 'DateRevised'], ['BookDocument', 'DateRevised']]},

  'path_journal_title': {'paths': [['MedlineCitation', 'Article', 'Journal', 'Title']]},
  'path_journal_iso': {'paths': [['MedlineCitation', 'Article', 'Journal', 'ISOAbbreviation'], ['MedlineCitation', 'MedlineJournalInfo', 'MedlineTA']]},
  'path_journal_issn': {'paths': [['MedlineCitation', 'Article', 'Journal', 'ISSN'], ['MedlineCitation', 'MedlineJournalInfo', 'ISSNLinking']]},

  'path_keywords': {'paths': [['MedlineCitation', 'KeywordList']]}
}

IGNORE_NO_ITEMS = True

SILENT = True


from xml.dom import minidom

class ReadPMC(object):
  dom_data = None

  def read_xml(self, xml):
    self.dom_data = minidom.parse(xml)

  def get_pmids(self, xml):
    self.read_xml(xml)

    pmid_list = []

    # Iterate the documents
    for doc in self.dom_data.getElementsByTagName('article'):
      # Retrieve the pmids
      pmid_list.append(self.get_pmid(doc))

    cleaned_pmid_list = list(filter(None.__ne__, pmid_list))

    list_len = len(pmid_list)
    clean_len = len(cleaned_pmid_list)

    if list_len - clean_len > 0:
      print('Removed %i items from the list because no pubmed id could be retrieved.' % (list_len - clean_len))

    return cleaned_pmid_list

  def get_pmid(self, article):
    for item in article.getElementsByTagName('article-id'):
      if item.attributes['pub-id-type'].value == 'pmid':
        # if item.firstChild.nodeValue is None:
        #   article.toprettyxml()

        return item.firstChild.nodeValue


from setuptools import setup

setup(name='PmedConnect',
      version='0.1.3',
      description='Searches and fetches results from PubMed',
      url='http://github.com/AMCeScience/PmedConnect',
      author='Allard van Altena',
      author_email='a.j.vanaltena@amc.uva.nl',
      license='GPL-3.0',
      packages=['PmedConnect'],
      install_requires=[
        'progressbar2',
        'Biopython'
      ],
      test_suite='nose.collector',
    	tests_require=[
    		'nose'
  		],
      include_package_data=True,
      zip_safe=False)